<Query Kind="Program">
  <Connection>
    <ID>ad2274e5-21dc-4e2c-9f3e-e55feb544557</ID>
    <Driver Assembly="StreamInsightLinqPad" PublicKeyToken="3d3a4b0768c9178e">StreamInsightLinqPad.StaticDriver</Driver>
    <CustomAssemblyPath>StreamInsightLinqPad.Samples.dll</CustomAssemblyPath>
    <CustomTypeName>StreamInsightLinqPad.Samples.StreamInsightContext</CustomTypeName>
    <DriverData>
      <ContextKind>StreamInsight</ContextKind>
      <RequiredVersion>21.0.0.0</RequiredVersion>
    </DriverData>
  </Connection>
</Query>

void Main()
{
    ////////////////////////////////////////////////////////////////////////////////////////
    // How do I join two streams of data?
    // - StreamInsight events must:
    // (1) Overlap in time
    // (2) Satisfy the join condition
    // in order to join
    ////////////////////////////////////////////////////////////////////////////////////////
    
    // Create a stream of events, containing successive integers
    var now = DateTime.Now;
    var input1 = Application
        .DefineEnumerable(() => Enumerable.Range(1, 5))
        .ToPointStreamable(
            i => PointEvent.CreateInsert(now.AddSeconds(i), i),
            AdvanceTimeSettings.StrictlyIncreasingStartTime);
    
    var input2 = Application
        .DefineEnumerable(() => Enumerable.Range(1, 5))
        .ToPointStreamable(
            i => PointEvent.CreateInsert(now.AddSeconds(i).AddMilliseconds(1), i),
            AdvanceTimeSettings.StrictlyIncreasingStartTime);
    
    // Let's look at the two streams - note that the start times do not overlap
    DisplayPoints(input1, "Input Stream 1");
    DisplayPoints(input2, "Input Stream 2");
      
    // Join the two streams based on their value field being equal.  This will 
    // produce no results, as the events do not overlap in time.
    var joinNoResults = from i1 in input1
                        join i2 in input2
                        on i1 equals i2
                        select i1;
    
    DisplayIntervals(joinNoResults, "Join without events overlapping in time");
    
    // To have the events overlap in time, let's extend the duration of the first 
    // stream's events so that events are overlapping.
    var equiJoin = from i1 in input1.AlterEventDuration(e => TimeSpan.FromMilliseconds(10))
                   join i2 in input2 
                   on i1 equals i2
                   select i1;
    
    DisplayIntervals(equiJoin, "Join with time overlap");
      
    ///////////////////////////////////////////////////////////////////////////////////////
    // How do I join more than two streams?
    ///////////////////////////////////////////////////////////////////////////////////////
    var moreData = Application.DefineEnumerable(() => new[] {
        new { Key = 1, Description = "One Tag Event" },
        new { Key = 2, Description = "Two Tag Event" }, 
        new { Key = 3, Description = "Three Tag Event" }, 
        new { Key = 4, Description = "Four Tag Event" }, 
        new { Key = 5, Description = "Five Tag Event" },	
    });
    var moreStream = moreData.ToIntervalStreamable(
        p => IntervalEvent.CreateInsert(now, now.AddDays(1), p),
        AdvanceTimeSettings.IncreasingStartTime);
      
    // Joining three streams - join two streams, then again join two streams
    var threeWayJoin = from i1 in input1.AlterEventDuration(e => TimeSpan.FromMilliseconds(10))
                       join i2 in input2 on i1 equals i2
                       join i3 in moreStream on i1 equals i3.Key
                       select new
                       {
                           Value = i1,
                           Description = i3.Description
                       };
            
    DisplayIntervals(threeWayJoin, "Join with three streams");
}

static void DisplayPoints<T>(IQStreamable<T> query, string description)
{
    (from i in query.ToIntervalEnumerable()
     where i.EventKind == EventKind.Insert
     select new { StartTime = i.StartTime.ToString("hh:mm:ss.fff"), i.Payload }).Dump(description);
}

static void DisplayIntervals<T>(IQStreamable<T> query, string description)
{
    (from i in query.ToIntervalEnumerable()
     where i.EventKind == EventKind.Insert
     select i).Dump(description);
}