<Query Kind="Statements">
  <Connection>
    <ID>ad2274e5-21dc-4e2c-9f3e-e55feb544557</ID>
    <Driver Assembly="StreamInsightLinqPad" PublicKeyToken="3d3a4b0768c9178e">StreamInsightLinqPad.StaticDriver</Driver>
    <CustomAssemblyPath>StreamInsightLinqPad.Samples.dll</CustomAssemblyPath>
    <CustomTypeName>StreamInsightLinqPad.Samples.StreamInsightContext</CustomTypeName>
    <DriverData>
      <ContextKind>StreamInsight</ContextKind>
      <RequiredVersion>21.0.0.0</RequiredVersion>
    </DriverData>
  </Connection>
</Query>

//////////////////////////////////////////////////////////////////////////////////
// How do I detect if event A is followed by event B within 90 seconds?
//////////////////////////////////////////////////////////////////////////////////
var sourceDataAB = Application.DefineEnumerable(() => new []
{
    new { SourceId = "A", Value = 22, Status = 1, TimeStamp = new DateTime(2009, 10, 23, 16, 12, 0, DateTimeKind.Utc) },
    new { SourceId = "A", Value = 24, Status = 0, TimeStamp = new DateTime(2009, 10, 23, 16, 13, 0, DateTimeKind.Utc) },
    new { SourceId = "A", Value = 31, Status = 1, TimeStamp = new DateTime(2009, 10, 23, 16, 14, 0, DateTimeKind.Utc) },
    new { SourceId = "A", Value = 67, Status = 0, TimeStamp = new DateTime(2009, 10, 23, 16, 15, 0, DateTimeKind.Utc) },
    new { SourceId = "A", Value = 54, Status = 0, TimeStamp = new DateTime(2009, 10, 23, 16, 16, 0, DateTimeKind.Utc) },
    new { SourceId = "A", Value = 50, Status = 1, TimeStamp = new DateTime(2009, 10, 23, 16, 30, 0, DateTimeKind.Utc) },
    new { SourceId = "A", Value = 87, Status = 0, TimeStamp = new DateTime(2009, 10, 23, 16, 35, 0, DateTimeKind.Utc) },
});

var sourceAB = sourceDataAB.ToPointStreamable(
    ev => PointEvent.CreateInsert(ev.TimeStamp.ToLocalTime(), ev),
    AdvanceTimeSettings.StrictlyIncreasingStartTime);

// Here, every event is considered to be event A, while event B is defined
// as having a value that increased by more than 30 regarding the previous value.
var resultAB = from first in sourceAB.AlterEventDuration(e => TimeSpan.FromSeconds(90))
               join second in sourceAB 
               on first.SourceId equals second.SourceId
               where second.Value > first.Value + 30
               select new
               {
                   second.SourceId,
                   second.Value,
                   delta = second.Value - first.Value
               };
        
(from e in resultAB.ToIntervalEnumerable()
 where e.EventKind == EventKind.Insert
 select e).Dump("A and B occuring within 5 minutes of each other");

//////////////////////////////////////////////////////////////////////////////////
// How do I detect if event A is not followed by event B within 5 minutes?
// ------------------------------------------------------------------------------
// Producing an event when a specific input event A is not followed by
// another specific input event B within a certain period.
//////////////////////////////////////////////////////////////////////////////////

var sourceDataNoAB = Application.DefineEnumerable(() => new []
{
    new { SourceId = "A", Value = 0, Status = 0, TimeStamp = new DateTime(2009, 10, 23, 16, 0, 0, DateTimeKind.Utc) },
    new { SourceId = "A", Value = 0, Status = 0, TimeStamp = new DateTime(2009, 10, 23, 16, 10, 0, DateTimeKind.Utc) },
    new { SourceId = "A", Value = 0, Status = 1, TimeStamp = new DateTime(2009, 10, 23, 16, 12, 0, DateTimeKind.Utc) },
    new { SourceId = "A", Value = 0, Status = 0, TimeStamp = new DateTime(2009, 10, 23, 16, 13, 0, DateTimeKind.Utc) },
    new { SourceId = "A", Value = 0, Status = 1, TimeStamp = new DateTime(2009, 10, 23, 16, 14, 0, DateTimeKind.Utc) },
    new { SourceId = "A", Value = 0, Status = 0, TimeStamp = new DateTime(2009, 10, 23, 16, 15, 0, DateTimeKind.Utc) },
    new { SourceId = "A", Value = 0, Status = 1, TimeStamp = new DateTime(2009, 10, 23, 16, 20, 0, DateTimeKind.Utc) },
    new { SourceId = "A", Value = 0, Status = 0, TimeStamp = new DateTime(2009, 10, 23, 16, 30, 0, DateTimeKind.Utc) },
    new { SourceId = "A", Value = 0, Status = 0, TimeStamp = new DateTime(2009, 10, 23, 16, 35, 0, DateTimeKind.Utc) },
});

var sourceNoAB = sourceDataNoAB.ToPointStreamable(
    ev => PointEvent.CreateInsert(ev.TimeStamp.ToLocalTime(), ev),
    AdvanceTimeSettings.StrictlyIncreasingStartTime);

// Here, event A is defined as having a status of 1. Move all events
// A to the point where the potential timeout would occur.
var timedOut = (from e in sourceNoAB
                where e.Status == 1
                select e).ShiftEventTime(e => TimeSpan.FromMinutes(5));

// Here, event B is defined as having a status of 0. Extend all events
// B by the timeout.
var following = (from e in sourceNoAB
                 where e.Status == 0
                 select e).AlterEventDuration(e => TimeSpan.FromMinutes(5));

// Mask away the potential timeout events by the extended B events.
// - If B did not occur within the timeout, then the shifted A event
// will remain unchanged by the left-anti-semi-join and represent
// the desired output.
var resultNoAB = from t in timedOut.LeftAntiJoin(following, t => t.SourceId, c => c.SourceId)
                 select new { alarmMsg = "Status has not been reset in time!", t.SourceId };

(from p in resultNoAB.ToPointEnumerable()
 where p.EventKind == EventKind.Insert
 select p).Dump("A not followed by B within 5 minutes");

//////////////////////////////////////////////////////////////////////////////////
// How do I detect if A, B and C occur within 5 minutes of each other?
//////////////////////////////////////////////////////////////////////////////////

// Create a basic stream of events.  For this example we are looking for an instance
// where (2) (3) & (4) happen within 5 minutes of each other.
var data = Application.DefineEnumerable(() => Enumerable.Range(1, 10));

var inputStream = data.ToPointStreamable(
    payload => PointEvent.CreateInsert(DateTime.Now + TimeSpan.FromMinutes(payload), new { payload }), 
    AdvanceTimeSettings.IncreasingStartTime);

// Stretch out the input stream to our desired window size (5 minutes).  Join the
// stream against itself to create a synthetic event where events (2) and (3) are
// present at the same time (i.e. within the 5 minute window).
var selfJoin = from e1 in inputStream.AlterEventDuration(e => TimeSpan.FromMinutes(5))
               from e2 in inputStream.AlterEventDuration(e => TimeSpan.FromMinutes(5))
               where e1.payload == 2 && e2.payload == 3
               select new 
               {
                   a = e1.payload,
                   b = e2.payload
               };

// Next join the (2) (3) stream against the (4) stream to complete the pattern match
var selfJoin2 = from e3 in inputStream.AlterEventDuration(e => TimeSpan.FromMinutes(5))
                from e1 in selfJoin
                where e1.a == 2 && e1.b == 3 && e3.payload == 4
select new
{
a = e1.a,
b = e1.b,
c = e3.payload
};

var result = selfJoin2.ToIntervalEnumerable()
.Where (j => j.EventKind == EventKind.Insert )
.Dump("A, B and C occuring within 5 minutes of each other");

//////////////////////////////////////////////////////////////////////////////////
// How do I detect if A, B and C do *not* occur within 5 minutes of each other?
// -------------------------------------------------------------------------------
// Assume that detection is started on the occurence of A. I.e. Given point events {A,B,C}
// which occur disjointly in a 5 minute window,  the sequence (A,B,C) would not trigger an
// event, whereas the sequence (B,A,C) would (since event B is never matched, as it occurs
// before A).
//////////////////////////////////////////////////////////////////////////////////
var sourceData = Application.DefineEnumerable(() => new []
{
    new { StartTime = new DateTime(2009, 6, 25, 0, 00, 00, DateTimeKind.Utc), ID = "B"},
    new { StartTime = new DateTime(2009, 6, 25, 0, 00, 01, DateTimeKind.Utc), ID = "C"},
    new { StartTime = new DateTime(2009, 6, 25, 0, 00, 02, DateTimeKind.Utc), ID = "A"}, // event

    new { StartTime = new DateTime(2009, 6, 25, 0, 05, 00, DateTimeKind.Utc), ID = "A"}, // event

    new { StartTime = new DateTime(2009, 6, 25, 0, 11, 00, DateTimeKind.Utc), ID = "A"},
    new { StartTime = new DateTime(2009, 6, 25, 0, 11, 00, DateTimeKind.Utc), ID = "B"},
    new { StartTime = new DateTime(2009, 6, 25, 0, 11, 00, DateTimeKind.Utc), ID = "C"},

    new { StartTime = new DateTime(2009, 6, 25, 0, 15, 59, DateTimeKind.Utc), ID = "A"}, // event
    new { StartTime = new DateTime(2009, 6, 25, 0, 15, 59, DateTimeKind.Utc), ID = "A"}, // event
    new { StartTime = new DateTime(2009, 6, 25, 0, 16, 00, DateTimeKind.Utc), ID = "A"},
    new { StartTime = new DateTime(2009, 6, 25, 0, 16, 00, DateTimeKind.Utc), ID = "A"},
    new { StartTime = new DateTime(2009, 6, 25, 0, 18, 00, DateTimeKind.Utc), ID = "B"},
    new { StartTime = new DateTime(2009, 6, 25, 0, 20, 59, DateTimeKind.Utc), ID = "C"},

    new { StartTime = new DateTime(2009, 6, 25, 0, 25, 59, DateTimeKind.Utc), ID = "A"}, // event
    new { StartTime = new DateTime(2009, 6, 25, 0, 26, 01, DateTimeKind.Utc), ID = "B"},

    new { StartTime = new DateTime(2009, 6, 25, 0, 29, 59, DateTimeKind.Utc), ID = "A"}, // event
    new { StartTime = new DateTime(2009, 6, 25, 0, 30, 59, DateTimeKind.Utc), ID = "C"},
});

var source = sourceData.ToPointStreamable(
    ev => PointEvent.CreateInsert(ev.StartTime.ToLocalTime(), ev),
    AdvanceTimeSettings.IncreasingStartTime);

// Set up the window size and a small offset
var windowSize = TimeSpan.FromMinutes(5);
var oneTick = new TimeSpan(1);

// Filter input so we have a dedicated stream for each event ID of interest
var aStream = from e in source where e.ID == "A" select e;
var bStream = from e in source where e.ID == "B" select e;
var cStream = from e in source where e.ID == "C" select e;

// Use join to find when event A is followed by B within 5 minutes. This is accomplished by shifting
// the start time of B back by 5 minutes*, extending its duration by 5 minutes, and joining with A
var abStream = from a in aStream
               // *Note: because event lifetimes are closed on the left and open on the right,
               // when we alter the lifetime of B we need to add one tick to ensure
               // the original StartTime is included in the lifetime of the new event
               // (and subsequently, the join result)
               from b in bStream.AlterEventLifetime(e => e.StartTime + oneTick - windowSize, e => windowSize)
               where true
               select a;

// Join cStream with previous result (abStream) to find when A is followed by B and C within 5 minutes
var abcStream = from a in abStream
                from c in cStream.AlterEventLifetime(e => e.StartTime + oneTick - windowSize, e => windowSize)
                select a;

// Left Anti-Semi-Join occurences of A (aStream) with occurences of A followed by B and C within 5 minutes (abcStream)
// to find occurences of A which is not followed by B and C within 5 minutes
var result2 = aStream.LeftAntiJoin(abcStream, (a, abc) => true);

(from e in result2.ToIntervalEnumerable()
 where e.EventKind == EventKind.Insert
 select e).Dump("How do I detect if A, B and C do *not* occur within 5 minutes of each other?");