<Query Kind="Statements">
  <Connection>
    <ID>ad2274e5-21dc-4e2c-9f3e-e55feb544557</ID>
    <Driver Assembly="StreamInsightLinqPad" PublicKeyToken="3d3a4b0768c9178e">StreamInsightLinqPad.StaticDriver</Driver>
    <CustomAssemblyPath>StreamInsightLinqPad.Samples.dll</CustomAssemblyPath>
    <CustomTypeName>StreamInsightLinqPad.Samples.StreamInsightContext</CustomTypeName>
    <DriverData>
      <ContextKind>StreamInsight</ContextKind>
      <RequiredVersion>21.0.0.0</RequiredVersion>
    </DriverData>
  </Connection>
</Query>

// Create some basic sample data
var sourceData = Application.DefineEnumerable(() => new []
{
    new { SourceId = "A", Value = 22, Status = "SUCCESS", TimeStamp = new DateTime(2009, 10, 23, 16, 12, 0, DateTimeKind.Utc) },
    new { SourceId = "A", Value = 24, Status = "FAIL", TimeStamp = new DateTime(2009, 10, 23, 16, 12, 1, DateTimeKind.Utc) },
    new { SourceId = "A", Value = 31, Status = "SUCCESS", TimeStamp = new DateTime(2009, 10, 23, 16, 12, 3, DateTimeKind.Utc) },
    new { SourceId = "A", Value = 67, Status = "FAIL", TimeStamp = new DateTime(2009, 10, 23, 16, 13, 0, DateTimeKind.Utc) },
    new { SourceId = "A", Value = 54, Status = "FAIL", TimeStamp = new DateTime(2009, 10, 23, 16, 13, 2, DateTimeKind.Utc) },
    new { SourceId = "A", Value = 50, Status = "SUCCESS", TimeStamp = new DateTime(2009, 10, 23, 16, 14, 0, DateTimeKind.Utc) },
    new { SourceId = "A", Value = 87, Status = "FAIL", TimeStamp = new DateTime(2009, 10, 23, 16, 14, 1, DateTimeKind.Utc) },
});

var source = sourceData.ToPointStreamable(
    ev => PointEvent.CreateInsert(ev.TimeStamp.ToLocalTime(), ev),
    AdvanceTimeSettings.StrictlyIncreasingStartTime);

////////////////////////////////////////////////////////////////////////////////////////
// How do I calculate the percentage of successful events (or events matching a 
// condition) within a 2 minute window?
////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
// Approach One: 
// - Count the total number of events in each window
// - Count matching events in each window
// - Compute the percentage by joining the results
//////////////////////////////////////////////////////////////////////////////////////////

// Count total number of events in each 2 min window
var countAllEvents = from win in source.HoppingWindow(TimeSpan.FromMinutes(2), TimeSpan.FromMinutes(1))
                     select win.Count();

// Count of matching events *only* in each 2 min window
var countSuccessfulEvents = from win in (from s in source
                                         // Specify the "match" condition
                                         where s.Status == "SUCCESS" 
                                         select s).HoppingWindow(TimeSpan.FromMinutes(2), TimeSpan.FromMinutes(1))
                            select win.Count();

// Compute percentage of successful events. 
var percentOfSuccessfulEvents = 
    from all in countAllEvents
    from successful in countSuccessfulEvents
    select (double)successful / (double)all * 100.0;

(from e in percentOfSuccessfulEvents.ToPointEnumerable()
 where e.EventKind == EventKind.Insert
 select e).Dump("Percentage of success for 2 minute window - Approach One");
  
//////////////////////////////////////////////////////////////////////////////////////////
// Approach Two (avoiding the join):  
// - Convert the match condition into a numeric field in the stream
// - Sum against the numeric field to indicate count of success events
//////////////////////////////////////////////////////////////////////////////////////////
var annotatedSource = from ev in source 
                      select new 
                      { 
                          ev.SourceId, 
                          ev.Value, 
                          Success = (ev.Status == "SUCCESS") ? 1 : 0 
                      };

var ratios = from win in annotatedSource.HoppingWindow(TimeSpan.FromMinutes(2), TimeSpan.FromMinutes(1))
             select new 
             {
                CountSuccessful = win.Sum(e => e.Success), 
                CountAll = win.Count()
             };

var percentOfSuccessfulEvents2 = from e in ratios		
                                 select (double)e.CountAll / (double)e.CountSuccessful * 100D;
  
(from e in percentOfSuccessfulEvents.ToPointEnumerable()
 where e.EventKind == EventKind.Insert
 select e).Dump("Percentage via annotated field - Approach Two");