<Query Kind="Statements">
  <Connection>
    <ID>ad2274e5-21dc-4e2c-9f3e-e55feb544557</ID>
    <Driver Assembly="StreamInsightLinqPad" PublicKeyToken="3d3a4b0768c9178e">StreamInsightLinqPad.StaticDriver</Driver>
    <CustomAssemblyPath>StreamInsightLinqPad.Samples.dll</CustomAssemblyPath>
    <CustomTypeName>StreamInsightLinqPad.Samples.StreamInsightContext</CustomTypeName>
    <DriverData>
      <ContextKind>StreamInsight</ContextKind>
      <RequiredVersion>21.0.0.0</RequiredVersion>
    </DriverData>
  </Connection>
</Query>

////////////////////////////////////////////////////////////////////////////////////////////////
// How do I ignore transient conditions?  
// ---------------------------------------------------------------------------------------
// That is, given an "alarm" value for a given event, ignore the event if the value changes
// within 5 seconds (i.e. raise an event when a value does not change for a given time period).
////////////////////////////////////////////////////////////////////////////////////////////////

// Assuming a timeout of 2 minutes, only the alarm event (status 1) at 4:30
// is a true alarm, while the alarms at 4:12 and 4:14 are transient.
var sourceData = Application.DefineEnumerable(() => new []
{
    new { Status = 1, TimeStamp = new DateTime(2009, 10, 23, 16, 12, 0, DateTimeKind.Utc) },
    new { Status = 0, TimeStamp = new DateTime(2009, 10, 23, 16, 13, 0, DateTimeKind.Utc) },
    new { Status = 1, TimeStamp = new DateTime(2009, 10, 23, 16, 14, 0, DateTimeKind.Utc) },
    new { Status = 0, TimeStamp = new DateTime(2009, 10, 23, 16, 15, 0, DateTimeKind.Utc) },
    new { Status = 0, TimeStamp = new DateTime(2009, 10, 23, 16, 16, 0, DateTimeKind.Utc) },
    new { Status = 1, TimeStamp = new DateTime(2009, 10, 23, 16, 30, 0, DateTimeKind.Utc) },
    new { Status = 0, TimeStamp = new DateTime(2009, 10, 23, 16, 35, 0, DateTimeKind.Utc) },
});

var source = sourceData.ToPointStreamable(
    ev => PointEvent.CreateInsert(ev.TimeStamp.ToLocalTime(), ev),
    AdvanceTimeSettings.StrictlyIncreasingStartTime);

var timeout = TimeSpan.FromMinutes(2);

// take all alarm events that are not followed by a non-alarm event
// within the timeout
var result = source.Where(e => e.Status == 1).LeftAntiJoin(
    from n in source.AlterEventLifetime(n => n.StartTime.Subtract(timeout), n => timeout)
    where n.Status == 0
    select n, (e, n) => true);

(from p in result.ToIntervalEnumerable()
 where p.EventKind == EventKind.Insert
 select p).Dump("Ignore transient conditions");
    
////////////////////////////////////////////////////////////////////////////////////////////////		
// How do I ignore successive conditions (How do I prevent alarm floods?
// ------------------------------------------------------------------------------------------
// That is, given an alarm condition (an event), how do ignore successive alarm events within 
// a given time window (i.e. I know the garage door is open, stop telling me)."

// Status 0 encodes normal, Status 1 encodes an alarm.
// - With a timeout of 10min we expect the alarms at 4:12 and 4:35 to be
//   reported and all others to be filtered out.
var sourceData2 = Application.DefineEnumerable(() => new []
{
    new { Status = 0, TimeStamp = new DateTime(2009, 10, 23, 16, 0, 0, DateTimeKind.Utc) },
    new { Status = 0, TimeStamp = new DateTime(2009, 10, 23, 16, 10, 0, DateTimeKind.Utc) },
    new { Status = 1, TimeStamp = new DateTime(2009, 10, 23, 16, 12, 0, DateTimeKind.Utc) },
    new { Status = 1, TimeStamp = new DateTime(2009, 10, 23, 16, 13, 0, DateTimeKind.Utc) },
    new { Status = 1, TimeStamp = new DateTime(2009, 10, 23, 16, 14, 0, DateTimeKind.Utc) },
    new { Status = 1, TimeStamp = new DateTime(2009, 10, 23, 16, 15, 0, DateTimeKind.Utc) },
    new { Status = 0, TimeStamp = new DateTime(2009, 10, 23, 16, 20, 0, DateTimeKind.Utc) },
    new { Status = 0, TimeStamp = new DateTime(2009, 10, 23, 16, 30, 0, DateTimeKind.Utc) },
    new { Status = 1, TimeStamp = new DateTime(2009, 10, 23, 16, 35, 0, DateTimeKind.Utc) },
});

var source2 = sourceData2.ToPointStreamable(
    ev => PointEvent.CreateInsert(ev.TimeStamp.ToLocalTime(), ev),
    AdvanceTimeSettings.StrictlyIncreasingStartTime);

var timeout2 = TimeSpan.FromMinutes(10);

// Expand all alarm events to the timeout and count over snapshots
var counts = from win in source
                    .Where(e => e.Status == 1)
                    .AlterEventDuration(e => timeout2)
                    .SnapshotWindow()
             select new { count = win.Count() };

// Those snapshots with a count of 1 belong to the initial alarms.
// reduce to points and join with original stream.
var result2 = from c in counts
                    .Where(e => e.count == 1)
                    .ToPointEventStream()
              from e in source
              select e;

(from p in result2.ToIntervalEnumerable()
 where p.EventKind == EventKind.Insert
 select p).Dump("Ignore alarm floods / successive events");