<Query Kind="Program">
  <Connection>
    <ID>ad2274e5-21dc-4e2c-9f3e-e55feb544557</ID>
    <Driver Assembly="StreamInsightLinqPad" PublicKeyToken="3d3a4b0768c9178e">StreamInsightLinqPad.StaticDriver</Driver>
    <CustomAssemblyPath>StreamInsightLinqPad.Samples.dll</CustomAssemblyPath>
    <CustomTypeName>StreamInsightLinqPad.Samples.StreamInsightContext</CustomTypeName>
    <DriverData>
      <ContextKind>StreamInsight</ContextKind>
      <RequiredVersion>21.0.0.0</RequiredVersion>
    </DriverData>
  </Connection>
  <GACReference>Microsoft.SqlServer.Types, Version=11.0.0.0, Culture=neutral, PublicKeyToken=89845dcd8080cc91</GACReference>
  <Namespace>Microsoft.SqlServer.Types</Namespace>
</Query>

void Main()
{
    ////////////////////////////////////////////////////////////////////////////////////////////
    // Given location information (GPS, coordinates, etc) how do I detect if event A and 
    // event B happen within 5 miles of each other in the same 5 minute window?
    // 
    // How can I use the SQL Geography libraries in StreamInsight queries? The SQL geography
    // libraries are installed with SQL Server 2012.
    //
    // For clarity, assume we want to detect all pairs of distinct events which are 
    // within 5 miles of each other in a 5 minute window. / One solution is to:
    // (1) Use AlterEventDuration to extend the lifetime of an event to the 
    // requisite 5 minutes 
    // (2) Then use a cross join on the original input stream.  
    //
    // This obtains the set of event pairs which occur in the same 5 minute window. 
    // We then complete this exercise by filtering the result to discard self matches 
    // and event pairs which are greater than 5 miles away.
    ////////////////////////////////////////////////////////////////////////////////////////////	
    
    // Create a set of reference data with the locations of various assets
    var assetData = Application.DefineEnumerable(() => new []
    {
        new { AssetId = 1, AssetName = "My Stuff", 		 Latitude = 47.64339, Longitude = -122.12840 }, 
        new { AssetId = 2, AssetName = "My Other Stuff", Latitude = 47.64827, Longitude = -122.13171 } 
    });
    
    var now = DateTime.Now;
    
    var liveData = Application.DefineEnumerable(() => new []
    {
        new { Timestamp = now.AddMinutes(1),  AssetId = 1, State = "Green", Id = 1 },
        new { Timestamp = now.AddMinutes(2),  AssetId = 2, State = "Red", Id = 2 },
        new { Timestamp = now.AddMinutes(3),  AssetId = 1, State = "Green", Id = 3 },
        new { Timestamp = now.AddMinutes(4),  AssetId = 2, State = "Green", Id = 4  }	
    });
  
    // Convert the asset data into a reference stream
    var assetStream = assetData.ToPointStreamable(
        t => PointEvent.CreateInsert(now, t),
        AdvanceTimeSettings.IncreasingStartTime);
    
    // Each asset reference event is valid for 7 days
    var assetRef = assetStream.AlterEventDuration(e => TimeSpan.FromDays(7));
    
    // Convert the live data into a live stream
    var liveStream = liveData.ToPointStreamable(
        t => PointEvent.CreateInsert(t.Timestamp, t),
        AdvanceTimeSettings.IncreasingStartTime);
    
    // Join the two streams to integrate the location information
    var joinedStream = from e1 in liveStream
                       join e2 in assetRef on e1.AssetId equals e2.AssetId
                       select new
                       {
                           AssetId = e1.AssetId,
                           Id = e1.Id,
                           State = e1.State,
                           Latitude = e2.Latitude,
                           Longitude = e2.Longitude
                       };
    
    // Cross join the streams in a 5 minute window
    var crossJoin = from left in joinedStream.AlterEventDuration(e => TimeSpan.FromMinutes(5))
                    from right in joinedStream 
                    where 
                        left.AssetId != right.AssetId &&	// Look for different asset types
                        left.State != right.State  &&		// Look for different state values
                        // Look for two events occuring within 5 miles of each other
                        GetDistance(left.Latitude, left.Longitude, right.Latitude, right.Longitude) <= 5.0
                    select new
                    {
                        AssetA = left.AssetId,
                        AssetB = right.AssetId,
                        EventIdA = left.Id,
                        EventIdB = right.Id,
                        StateA = left.State,
                        StateB = right.State,
                        Distance = GetDistance(left.Latitude, left.Longitude, right.Latitude, right.Longitude)
                    };

    (from p in crossJoin.ToIntervalEnumerable()
     where p.EventKind == EventKind.Insert
     select p).Dump("Identifying events with geo-location");						
}

const int SRID = 4326;
const double MetersPerMile = 1609.344;
    
/// <summary>
/// Returns the distance in miles, using the SqlGeography type
/// </summary>
public static double GetDistance(
    double xLatitude,
    double xLongitude,
    double yLatitude,
    double yLongitude)
{
    return SqlGeography.Point(xLatitude, xLongitude, SRID)
        .STDistance(SqlGeography.Point(yLatitude, yLongitude, SRID))
        .Value / MetersPerMile;
}