<Query Kind="Program">
  <Connection>
    <ID>8c50f33b-f431-47ce-ab08-85212ab1f0a3</ID>
    <Driver Assembly="StreamInsightLinqPad" PublicKeyToken="3d3a4b0768c9178e">StreamInsightLinqPad.StaticDriver</Driver>
    <CustomAssemblyPath>StreamInsightLinqPad.Samples.dll</CustomAssemblyPath>
    <CustomTypeName>StreamInsightLinqPad.Samples.StreamInsightContext</CustomTypeName>
    <DriverData>
      <ContextKind>StreamInsight</ContextKind>
      <RequiredVersion>21.0.0.0</RequiredVersion>
    </DriverData>
  </Connection>
</Query>

// This sample demonstrates the design and usage of LINQ macros.
//
// Not Followed by is called with two input streams. When no event in
// the second stream occurs within the specified timeout after an event in the
// first stream, matching the first event through the given condition, a
// result event is returned by the macro.
//
// With the example source data, such a condition is met for stream A at
// input time 25, since the next reset event follows only at 30.
void Main()
{
    var startTime = new DateTimeOffset(2011, 1, 11, 8, 0, 0, TimeSpan.Zero);
    
    var sourceData = Application.DefineEnumerable(() => new []
    {
        new { Id = "A", Value = 12, Status = 0, TimeStamp = startTime.AddMinutes(0) },
        new { Id = "B", Value = 14, Status = 0, TimeStamp = startTime.AddMinutes(10) },
        new { Id = "A", Value = 21, Status = 1, TimeStamp = startTime.AddMinutes(12) },
        new { Id = "A", Value = 24, Status = 0, TimeStamp = startTime.AddMinutes(13) },
        new { Id = "B", Value = 22, Status = 1, TimeStamp = startTime.AddMinutes(14) },
        new { Id = "B", Value = 30, Status = 0, TimeStamp = startTime.AddMinutes(15) },
        new { Id = "A", Value = 12, Status = 1, TimeStamp = startTime.AddMinutes(20) },
        new { Id = "A", Value = 29, Status = 0, TimeStamp = startTime.AddMinutes(30) },
        new { Id = "B", Value = 18, Status = 0, TimeStamp = startTime.AddMinutes(35) },
    });
    
    var source = sourceData.ToPointStreamable(
        ev => PointEvent.CreateInsert(ev.TimeStamp, new { ev.Id, ev.Value, ev.Status }),
        AdvanceTimeSettings.StrictlyIncreasingStartTime);
    
    var streamA = from e in source where e.Status == 1 select e;
    var streamB = from e in source where e.Status == 0 select e;
    
    var result = NotFollowedBy(streamA, streamB, TimeSpan.FromMinutes(5), (a, b) => a.Id == b.Id);
    
    result.Dump("Events with status 1 not followed by status 0 (with the same Id) within 5 minutes (at the timeout):");
}

public static IQStreamable<TFirst> NotFollowedBy<TFirst, TSecond>(
    IQStreamable<TFirst> first,
    IQStreamable<TSecond> second,
    TimeSpan timeout,
    Expression<Func<TFirst, TSecond, bool>> predicate)
{
    if (first == null) throw new ArgumentNullException("first");
    if (second == null) throw new ArgumentNullException("second");
    if (predicate == null) throw new ArgumentNullException("predicate");
    
    // Move all events from first to the point where the potential timeout would occur.
    var timedOut = first
        .ShiftEventTime(_ => timeout);

    // Extend all events from second by the timeout.
    var following = second
        .AlterEventDuration(e => timeout);

    // mask away the potential timeout events by the extended second events.
    // If second event did not occur within the timeout, then the shifted first event
    // will remain unchanged by the left-anti-semi-join and represent
    // the desired output.
    return timedOut.LeftAntiJoin(following, predicate);
}