<Query Kind="Program">
  <Connection>
    <ID>8c50f33b-f431-47ce-ab08-85212ab1f0a3</ID>
    <Driver Assembly="StreamInsightLinqPad" PublicKeyToken="3d3a4b0768c9178e">StreamInsightLinqPad.StaticDriver</Driver>
    <CustomAssemblyPath>StreamInsightLinqPad.Samples.dll</CustomAssemblyPath>
    <CustomTypeName>StreamInsightLinqPad.Samples.StreamInsightContext</CustomTypeName>
    <DriverData>
      <ContextKind>StreamInsight</ContextKind>
      <RequiredVersion>21.0.0.0</RequiredVersion>
    </DriverData>
  </Connection>
</Query>

// This sample demonstrates how  a user-defined stream operator can be used
// to implement a calendar-based window.
//
// Built-in window operators only work with fixed durations, which is not
// practical for windowing over calendar months, since their length varies
// from 28 to 31 days. The UDSO in this sample computes the running sum
// within each month.
void Main()
{
    var startTime = DateTime.UtcNow;

    // generate random data over 120 days
    var random = new Random();
    var source = Application.DefineEnumerable(() => Enumerable.Range(0, 30).Select(i => PointEvent.CreateInsert(startTime.AddDays(i * 4), random.NextDouble()*20)));
    var input = source.ToStreamable(AdvanceTimeSettings.StrictlyIncreasingStartTime);

    var query = input.Scan(() => new MonthlyRunningSum());

    query.Dump();
}

[DataContract]
public class MonthlyRunningSum : Microsoft.ComplexEventProcessing.Extensibility.CepPointStreamOperator<double, double>
{
    [DataMember]
    double _runningSum;
    
    [DataMember]
    DateTimeOffset? _nextCti;

    public MonthlyRunningSum()
    {
        _runningSum = 0;
    }

    // As long as we didn't increment our sum, we don't have any state.
    public override bool IsEmpty
    {
        get { return _runningSum == 0; }
    }

    // Telling the engine when we want to see the next CTI.
    public override DateTimeOffset? NextCti
    {
        get { return _nextCti; }
    }


    public override IEnumerable<double> ProcessEvent(PointEvent<double> inputEvent)
    {
        // while we don't see the next CTI, we are yielding the running sum
        if (inputEvent.EventKind == EventKind.Insert)
        {
            // Event start time should be less than expected nextCTI
            System.Diagnostics.Debug.Assert(!_nextCti.HasValue || inputEvent.StartTime < _nextCti.Value);
      
            _runningSum += inputEvent.Payload;

            // first insert event after resetting
            if (!_nextCti.HasValue)
            {
                // we want to see a CTI when this month has ended
                _nextCti = GetBeginOfNextMonth(inputEvent.StartTime);
            }
            
            yield return _runningSum;
        }
        else
        {
            // we just got a CTI, so we reset
            _runningSum = 0;
            _nextCti = null;
        }
    }
  
    private DateTimeOffset? GetBeginOfNextMonth(DateTimeOffset dateTime)
    {
        if (dateTime.Month == 12)
        {
            return new DateTimeOffset(dateTime.Year + 1,
                                      1,
                                      1, 0, 0, 0, TimeSpan.Zero);
        }
        else
        {
            return new DateTimeOffset(dateTime.Year,
                                      dateTime.Month + 1,
                                      1, 0, 0, 0, TimeSpan.Zero);
        }
    }
}